#
# Copyright (c) 2023, HILSTER - https://hilster.io
# All rights reserved.
#

from htf.daqmx.core import DAQError
from htf.daqmx.task import Task
from ctypes import *  # noqa: F403
from htf.daqmx.api import (
    uInt32,
    int32,
    DAQmx_Val_ChanForAllLines,
    float64,
    DAQmx_Val_ChanPerLine,
)
from htf.mixins.delegation import DelegateMixin


class DigitalOutputSingleLine(DelegateMixin):
    """
    ``DigitalOutputSingleLine`` is used to access a single digital output line
    on a DAQmx device.

    A DAQmx task is automatically created and deleted in the constructor and
    destructor.
    """

    N = 2

    def __init__(self, descriptor: str) -> None:
        """
        Args:
            descriptor: The DAQmx device and line descriptor, eg.
                `cDaq1Mod1/line0`.
        """
        task_name = "DO_" + descriptor.replace("/", "_")
        self.state = 0x00000000
        self.data = cast((uInt32 * self.N)(), POINTER(uInt32))  # noqa: F405
        self.written = int32(0)
        self.read = int32(0)
        self.task = Task(task_name)

        if not isinstance(descriptor, bytes):
            descriptor = descriptor.encode()  # type: ignore

        self.task.CreateDOChan(descriptor, None, DAQmx_Val_ChanPerLine)

    def get(self) -> bool:
        """
        Read the current pin state.
        """
        self.task.ReadDigitalU32(
            self.N,
            float64(1.0),
            DAQmx_Val_ChanForAllLines,
            self.data,
            self.N,
            byref(self.read),  # noqa: F405
            None,
        )
        self.task.StopTask()
        return bool(self.data[0])

    def set(self, state: bool = True) -> None:
        """
        Set the pin state to ``state``.

        Args:
            state=True: new state for output pin.
        """
        if state:
            self.data[0] = uInt32(0xFFFFFFFF)
            self.data[1] = uInt32(0xFFFFFFFF)
        else:
            self.data[0] = uInt32(0x00000000)
            self.data[1] = uInt32(0x00000000)

        self.task.WriteDigitalU32(
            self.N,
            True,
            float64(1.0),
            DAQmx_Val_ChanForAllLines,
            self.data,
            byref(self.written),  # noqa: F405
            None,
        )
        self.task.StopTask()

    def reset(self) -> None:
        """
        Reset the pin state.
        """
        self.set(state=False)

    def close(self) -> None:
        """
        Close the abstractor by clearing the task.
        """
        try:
            self.task.ClearTask()
        except DAQError:
            pass

    def __del__(self) -> None:
        self.close()

    def __str__(self) -> str:
        return str(self.get())


if __name__ == "__main__":
    import time

    a = DigitalOutputSingleLine("Dev1/port0")
    b = DigitalOutputSingleLine("Dev1/port1")

    a.reset()
    assert a.get() is False
    assert b.get() is False

    a.set()
    assert a.get() is True
    assert b.get() is True

    b.reset()
    assert a.get() is False
    assert b.get() is False

    b.set()
    assert a.get() is True
    assert b.get() is True

    a.set()

    a.close()
    b.close()

    exit(0)
    c = DigitalOutputSingleLine("cDaq4Mod1/line0")

    assert c.get() is True

    c.close()

    # a.set()
    # b.set()
    # a.get()
    # b.get()

    exit(0)

    l7 = DigitalOutputSingleLine("cDaq4Mod1/line0")
    l8 = DigitalOutputSingleLine("cDaq4Mod1/line0")

    for ii in range(10):
        l7.reset()
        l8.set()
        time.sleep(0.5)

        l8.reset()
        l7.set()
        time.sleep(0.5)
        print(ii)
